import os
import time
import shutil

from glob import glob

import pytest

import requests_cache


def assert_sqlite(testdir, name=None, tmpdir=None, exists=True):
    name = name or ".pytest-check-links-cache.sqlite"
    tmpdir = str(tmpdir or testdir.tmpdir)
    caches = list(glob(os.path.join(tmpdir, name)))
    if exists:
        assert caches
    else:
        assert not caches


@pytest.mark.parametrize("cache_name", [
    None,
    "custom-cache"
])
def test_cache_expiry(testdir, base_args, cache_name, tmpdir):
    """will the default sqlite3 backend persist and then expire?
    """
    testdir.copy_example('linkcheck.ipynb')

    args = base_args + ["--check-links-cache-expire-after", "2"]
    if cache_name:
        args += ["--check-links-cache-name", os.path.join(str(tmpdir), cache_name)]
    expected = dict(passed=3, failed=3)
    t0 = time.time()
    result = testdir.runpytest(*args)
    t1 = time.time()
    result.assert_outcomes(**expected)

    if cache_name:
        assert_sqlite(testdir, name="{}.sqlite".format(cache_name), tmpdir=tmpdir)
    else:
        assert_sqlite(testdir)

    t2 = time.time()
    result = testdir.runpytest(*args)
    t3 = time.time()
    result.assert_outcomes(**expected)

    d0 = t1 - t0
    d1 = t3 - t2

    assert d0 > d1, "cache did not make second run faster"

    time.sleep(2)

    t4 = time.time()
    result = testdir.runpytest(*args)
    t5 = time.time()
    result.assert_outcomes(**expected)

    d2 = t5 - t4
    d3 = t3 - t2

    assert d2 > d3, "cache did not expire"


def test_cache_memory(testdir, memory_args):
    """will the memory backend cache links inside a run?
    """
    expected = dict(passed=3, failed=0)

    testdir.copy_example('httpbin.md')

    def run(passed):
        t0 = time.time()
        result = testdir.runpytest(*memory_args)
        t1 = time.time()
        result.assert_outcomes(passed=passed, failed=0)
        assert_sqlite(testdir, exists=False)
        return t1 - t0

    d0 = run(6)

    for i in range(5):
        shutil.copy(
            os.path.join(str(testdir.tmpdir), "httpbin.md"),
            os.path.join(str(testdir.tmpdir), "httpbin{}.md".format(i))
        )

    d1 = run(36)
    # allow a healthy savings margin for network flake
    assert d1 < d0 * 4


def test_cache_retry(testdir, memory_args):
    """will a Retry-After header work with cache?
    """

    testdir.copy_example('httpbin.md')

    attempts = []

    _get = requests_cache.CachedSession.get

    def mock_get(*args, **kwargs):
        response = _get(*args, **kwargs)
        if len(attempts) < 5:
            response.status_code = 502
            response.headers['Retry-After'] = '0'
        attempts.append([args, kwargs])
        return response

    requests_cache.CachedSession.get = mock_get

    result = testdir.runpytest(*memory_args)

    try:
        result.assert_outcomes(passed=5, failed=1)
        assert len(attempts) == 10
    finally:
        requests_cache.CachedSession.get = _get


def test_cache_backend_opts(testdir, base_args):
    testdir.copy_example('httpbin.md')
    args = base_args + [
        "--check-links-cache-backend-opt", "fast_save:true",
        "--check-links-cache-name", "foo",
        "--check-links-cache-backend-opt", "extension:.db",
        "--check-links-cache-backend-opt", "allowable_codes:[200]"
    ]
    result = testdir.runpytest(*args)
    result.assert_outcomes(passed=6, failed=0)
    assert_sqlite(testdir, name="foo.db")



from pathlib import Path

from .conftest import skip_pywin32


def test_ipynb(testdir):
    testdir.copy_example('linkcheck.ipynb')
    result = testdir.runpytest("-v", "--check-links")
    result.assert_outcomes(passed=3, failed=3)

def test_markdown(testdir):
    testdir.copy_example('markdown.md')
    result = testdir.runpytest("-v", "--check-links")
    result.assert_outcomes(passed=8, failed=4)
    result = testdir.runpytest("-v", "--check-links", "--check-links-ignore", "http.*example.com/.*")
    result.assert_outcomes(passed=8, failed=1)

def test_markdown_nested(testdir):
    testdir.copy_example('nested/nested.md')
    testdir.mkdir('nested')
    md = testdir.tmpdir / 'nested.md'
    md.move(testdir.tmpdir / 'nested' / 'nested.md')
    testdir.copy_example('markdown.md')
    result = testdir.runpytest("-v", "--check-links")
    result.assert_outcomes(passed=9, failed=4)
    result = testdir.runpytest("-v", "--check-links", "--check-links-ignore", "http.*example.com/.*")
    result.assert_outcomes(passed=9, failed=1)

@skip_pywin32
def test_rst(testdir):
    testdir.copy_example('rst.rst')
    result = testdir.runpytest("-v", "--check-links")
    result.assert_outcomes(passed=7, failed=2)

@skip_pywin32
def test_rst_nested(testdir):
    testdir.copy_example('nested/nested.rst')
    testdir.mkdir('nested')
    rst = testdir.tmpdir / 'nested.rst'
    rst.move(testdir.tmpdir / 'nested' / 'nested.rst')
    testdir.copy_example('rst.rst')
    result = testdir.runpytest("-v", "--check-links")
    result.assert_outcomes(passed=13, failed=5)

def test_link_ext(testdir):
    testdir.copy_example('linkcheck.ipynb')
    testdir.copy_example('rst.rst')
    testdir.copy_example('markdown.md')
    result = testdir.runpytest("-v", "--check-links", "--links-ext=md,ipynb")
    result.assert_outcomes(passed=11, failed=7)



import subprocess
import pytest


def test_cli_meta():
    assert subprocess.call(["pytest-check-links", "--version"]) == 0
    assert subprocess.call(["pytest-check-links", "--help"]) == 0


@pytest.mark.parametrize("example,rc,expected,unexpected", [
    ["httpbin.md", 0, [" 6 passed"], [" failed", " warning"]],
    ["rst.rst", 1, [" 2 failed", " 7 passed"], [" warning"]]
])
def test_cli_pass(testdir, example, rc, expected, unexpected):
    testdir.copy_example(example)
    testdir.copy_example("setup.cfg")
    proc = subprocess.Popen(["pytest-check-links"], stdout=subprocess.PIPE)
    stdout, stderr = proc.communicate()
    summary = stdout.decode('utf-8').strip().splitlines()[-1]
    assert rc == proc.returncode
    for ex in expected:
        assert ex in summary, stdout.decode('utf-8')
    for unex in unexpected:
        assert unex not in summary, stdout.decode('utf-8')



"""
This script will run pytest with our plugin enabled,
and with the normal python plugin disabled. This means collection
of normal tests will be disabled, unless other plugins are set
to collect further tests. If that is the case, you can disable
these plugins with the py.test command line option
"-p no:<plugin-name-here>".
"""
# pragma: no cover

import sys


def main(args=None):
    import pytest

    if args is None:
        args = sys.argv[1:]

    return pytest.main(args + [
        '--check-links',
    ], [
        'no:python',
    ])


if __name__ == '__main__':
    sys.exit(main())



import pytest


def test_anchors_local_self(testdir, anchor_args):
    testdir.copy_example('anchors_self.html')
    result = testdir.runpytest(*anchor_args)
    result.assert_outcomes(passed=2, failed=2)


def test_anchors_local_other(testdir, anchor_args):
    testdir.copy_example('anchors_self.html')
    testdir.copy_example('anchors_other.html')
    args = anchor_args + ["anchors_other.html"]
    result = testdir.runpytest(*args)
    result.assert_outcomes(passed=1, failed=2)


def test_anchors_external(testdir, anchor_args):
    testdir.copy_example('anchors_remote.html')
    result = testdir.runpytest(*anchor_args)
    result.assert_outcomes(passed=1, failed=1)



import argparse
import json

class StoreExtensionsAction(argparse.Action):

    def __init__(self, option_strings, dest, nargs=None, **kwargs):
        if nargs is not None:
            raise ValueError("nargs not allowed")
        super(StoreExtensionsAction, self).__init__(option_strings, dest, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        values = self.parse_extensions(values)
        setattr(namespace, self.dest, values)

    def parse_extensions(self, csv):
        return {'.%s' % ext.lstrip('.') for ext in csv.split(',')}


class StoreCacheAction(argparse.Action):
    """Build the cache session kwargs
    """
    def __call__(self, parser, namespace, values, option_string=None):
        ns_name = "check_links_cache_kwargs"
        if not hasattr(namespace, ns_name):
            setattr(namespace, ns_name, {})
        dest = self.dest.replace("check_links_cache_", "")
        kwargs = namespace.check_links_cache_kwargs
        if dest == "name":
            kwargs["cache_name"] = values
        elif dest == "expire_after":
            kwargs["expire_after"] = float(values)
        elif dest == "backend_opt":
            key, value = str(values).split(":", 1)
            try:
                kwargs[key] = json.loads(value)
            except:
                kwargs[key] = value
        else:
            kwargs[dest] = values



import sys
import pytest

pytest_plugins = ["pytester"]

skip_pywin32 = pytest.mark.skipif(sys.platform == "win32", reason="pywin32 double import")


@pytest.fixture
def anchor_args():
    return ["-v", "--check-links", "--check-anchors"]

@pytest.fixture
def base_args():
    return ["-v", "--check-links", "--check-links-cache"]


@pytest.fixture
def memory_args(base_args):
    return base_args + ["--check-links-cache-backend", "memory"]



#!/usr/bin/env python

from setuptools import setup

# This repository uses pbr - Python Build Reasonableness
# https://docs.openstack.org/pbr/latest/
# The version is set by the current git tag + describe status.
# The rest of the config data is in setup.cfg
setup(
    setup_requires=['pbr>=1.9', 'setuptools>=17.1'],
    pbr=True,
)



from docutils.core import publish_parts
import io
import os
import re
import time
import warnings

import html5lib
import pytest
from requests import Session, Request
from requests.compat import unquote

from .args import StoreExtensionsAction, StoreCacheAction

_ENC = 'utf8'

default_extensions = {'.md', '.rst', '.html', '.ipynb'}
supported_extensions = {'.md', '.rst', '.html', '.ipynb'}

default_cache = dict(
    cache_name='.pytest-check-links-cache',
    backend=None,
    expire_after=None,
    allowable_codes=list(range(200, 512)),
)


def pytest_addoption(parser):
    group = parser.getgroup("general")
    group.addoption('--check-links', action='store_true',
        help="Check links for validity")
    group.addoption('--check-anchors', action='store_true',
        help="Check link anchors for validity")
    group.addoption('--links-ext', action=StoreExtensionsAction,
        default=default_extensions,
        help="Which file extensions to check links for, "
             "as a comma-separated list of values. Supported "
             "extensions are: %s." %
                extensions_str(supported_extensions))
    group.addoption('--check-links-ignore', action='append',
        help="A list of regular expressions that match URIs that should not be checked.")
    group.addoption('--check-links-cache', action='store_true',
        help="Cache requests when checking links")
    group.addoption('--check-links-cache-name', action=StoreCacheAction,
        help="Name of link cache")
    group.addoption('--check-links-cache-backend', action=StoreCacheAction,
        help="Cache persistence backend")
    group.addoption('--check-links-cache-expire-after', action=StoreCacheAction,
        help="Time to cache link responses (seconds)")
    group.addoption('--check-links-cache-allowable-codes', action=StoreCacheAction,
        help="HTTP response codes to cache")
    group.addoption('--check-links-cache-backend-opt', action=StoreCacheAction,
        help="Backend-specific options for link cache, specfied as `opt:value`")



def pytest_configure(config):
    if config.option.links_ext:
        validate_extensions(config.option.links_ext)


def pytest_collect_file(path, parent):
    config = parent.config
    ignore_links = config.option.check_links_ignore
    if config.option.check_links:
        requests_session = ensure_requests_session(config)
        if path.ext.lower() in config.option.links_ext:
            check_anchors = config.option.check_anchors
            if hasattr(CheckLinks, "from_parent"):
                return CheckLinks.from_parent(parent, fspath=path, requests_session=requests_session, check_anchors=check_anchors, ignore_links=ignore_links)
            return CheckLinks(fspath=path, parent=parent, requests_session=requests_session, check_anchors=check_anchors, ignore_links=ignore_links)

def ensure_requests_session(config):
    """Build the singleton requests.Session (or subclass)
    """
    session_attr = "check_links_requests_session"

    if not hasattr(config.option, session_attr):
        if config.option.check_links_cache:
            from requests_cache import CachedSession
            conf_kwargs = getattr(config.option, "check_links_cache_kwargs", {})
            kwargs = dict(default_cache)
            kwargs.update(conf_kwargs)
            requests_session = CachedSession(**kwargs)
            if kwargs.get("expire_after"):
                requests_session.remove_expired_responses()
        else:
            requests_session = Session()

        requests_session.headers['User-Agent'] = 'pytest-check-links'

        setattr(config.option, session_attr, requests_session)

    return getattr(config.option, session_attr)


class CheckLinks(pytest.File):
    """Check the links in a file"""
    def __init__(self, parent=None, fspath=None, requests_session=None, check_anchors=False, ignore_links=None):
        super(CheckLinks, self).__init__(fspath, parent)
        self.check_anchors = check_anchors
        self.requests_session = requests_session
        self.ignore_links = ignore_links or []

    def _html_from_html(self):
        """Return HTML from an HTML file"""
        with io.open(str(self.fspath), encoding=_ENC) as f:
            return f.read()

    def _html_from_markdown(self):
        """Return HTML from a markdown file"""
        # FIXME: use commonmark or a pluggable engine
        from nbconvert.filters import markdown2html
        with io.open(str(self.fspath), encoding=_ENC) as f:
            markdown = f.read()
        return markdown2html(markdown)

    def _html_from_rst(self):
        """Return HTML from an rst file"""
        with io.open(str(self.fspath), encoding=_ENC) as f:
            rst = f.read()
        return publish_parts(rst, source_path=str(self.fspath), writer_name='html')['html_body']

    def _items_from_notebook(self):
        """Yield LinkItems from a notebook"""
        import nbformat
        from nbconvert.filters import markdown2html

        nb = nbformat.read(str(self.fspath), as_version=4)
        for cell_num, cell in enumerate(nb.cells):
            if cell.cell_type != 'markdown':
                continue

            html = markdown2html(cell.source)
            basename = 'Cell %i' % cell_num
            for item in links_in_html(basename, self, html):
                yield item

    def collect(self):
        path = self.fspath
        if path.ext == '.ipynb':
            for item in self._items_from_notebook():
                yield item
            return

        if path.ext == '.html':
            html = self._html_from_html()
        elif path.ext == '.md':
            html = self._html_from_markdown()
        elif path.ext == '.rst':
            html = self._html_from_rst()

        for item in links_in_html(path, self, html):
            ignore = False
            for pattern in self.ignore_links:
                if re.match(pattern, item.target):
                    ignore = True
            if not ignore:
                yield item


class BrokenLinkError(Exception):
    def __init__(self, url, error):
        self.url = url
        self.error = error

    def __repr__(self):
        return "<%s url=%s, error=%s>" % (
            self.__class__.__name__, self.url, self.error
        )


def links_in_html(base_name, parent, html):
    """Yield LinkItems from a markdown cell

    Parsed HTML with html5lib, yielding LinkItems for testing.
    """
    parsed = html5lib.parse(html, namespaceHTMLElements=False)

    for element in parsed.iter():
        url = None
        tag = element.tag
        if tag == 'a':
            attr = 'href'
            url = element.get('href', '')
            if url.startswith('#') and not parent.check_anchors:
                # skip internal links
                continue

        elif tag in {'img', 'iframe'}:
            attr = 'src'
        else:
            continue

        url = element.get(attr)
        name = '{} <{} {}={}>'.format(base_name, tag, attr, url)

        if url:
            if ':' in url:
                proto = url.split(':', 1)[0]
                if proto.lower() not in {'http', 'https'}:
                    # ignore non-http links (mailto:, data:, etc.)
                    continue
            if hasattr(LinkItem, "from_parent"):
                yield LinkItem.from_parent(parent, name=name, target=url, parsed=parsed)
            else:
                yield LinkItem(name=name, parent=parent, target=url, parsed=parsed)


class LinkItem(pytest.Item):
    """Test item for an HTML link

    Args:

        name, parent: inherited from pytest.Item
        target (str): The URL or path target for the link
        parsed (xml.etree.ElementTree.Element): The parsed HTML
        description (str, optional): The description to be used in the report header
    """
    def __init__(self, name=None, parent=None, target=None, parsed=None, description=''):
        super(LinkItem, self).__init__(name, parent)
        self.target = target
        self.parsed = parsed
        self.description = description or '{}: {}'.format(self.fspath, target)

    def repr_failure(self, excinfo):
        exc = excinfo.value
        if isinstance(exc, BrokenLinkError):
            return '{}: {}'.format(exc.url, exc.error)
        else:
            return super(LinkItem, self).repr_failure(excinfo)

    def reportinfo(self):
        return self.fspath, 0, self.description

    def sleep(self, headers):
        """Handle responses or errors with a Retry-After header.

        https://www.w3.org/Protocols/rfc2616/rfc2616-sec14.html#sec14.37
        """
        if headers is None:
            return False

        header = headers.get('Retry-After')

        if header is None:
            return False

        if header == '1m0s':
            sleep_time = 60
        else:
            try:
                sleep_time = int(header)
            except ValueError:
                sleep_time = 10

        time.sleep(sleep_time)

        return True

    def handle_anchor(self, parsed, anchor):
        """Verify an anchor exists in the parsed HTML
        """
        anchors = set(parsed.findall('*//a[@name="{}"]'.format(anchor)))
        anchors |= set(parsed.findall('*//*[@id="{}"]'.format(anchor)))

        if not anchors:
            raise BrokenLinkError(self.target, "Missing anchor: %s" % anchor)

        if len(anchors) > 1:
            raise BrokenLinkError(
                self.target,
                "Ambiguous anchor: %d (found %s)" % (
                    len(anchors), anchor
                )
            )

    def fetch_with_retries(self, url, retries=3):
        """Fetch a URL, optionally retrying after a delay (by header)
        """

        url_no_anchor = url.split("#")[0]
        session = self.parent.requests_session

        try:
            response = session.get(url_no_anchor)
        except Exception as err:
            if hasattr(err, 'headers') and retries and self.sleep(err.headers):
                self.uncache_url(url_no_anchor)
                return self.fetch_with_retries(url, retries=retries - 1)

            raise BrokenLinkError(url, "%s" % err)

        if response.status_code >= 400:
            if retries and self.sleep(response.headers):
                self.uncache_url(url_no_anchor)
                return self.fetch_with_retries(url, retries=retries - 1)

            raise BrokenLinkError(url, "%d: %s" % (
                response.status_code,
                response.reason
            ))

        return response

    def uncache_url(self, url):
        uncached = False
        session = self.parent.requests_session
        if hasattr(session, "cache"):
            request = Request('GET', url, headers=session.headers).prepare()
            key = session.cache.create_key(request)
            if session.cache.has_key(key):
                session.cache.delete(key)
                uncached = True
        return uncached

    def runtest(self):
        url = self.target

        if ':' in url:
            response = self.fetch_with_retries(url)
            if self.parent.check_anchors and '#' in url:
                anchor = url.split('#')[1]
                if anchor and "html" in response.headers.get("Content-Type"):
                    parsed = html5lib.parse(response.content, namespaceHTMLElements=False)
                    return self.handle_anchor(parsed, anchor)
        else:
            if url.startswith('/'):
                raise BrokenLinkError(url, "absolute path link")
            # relative URL
            anchor = None
            if '?' in url:
                url = url.split('?')[0]
            if '#' in url:
                url, anchor = url.split('#')

            if not url and anchor:
                if self.parent.check_anchors:
                    self.handle_anchor(self.parsed, anchor)
                return

            url_path = unquote(url).replace('/', os.path.sep)
            dirpath = self.fspath.dirpath()
            exists = False
            for ext in supported_extensions:
                rel_path = url_path.replace('.html', ext)
                target_path = dirpath.join(rel_path)
                if target_path.exists():
                    exists = True
                    # only check anchors in html for now
                    if ext == ".html" and anchor and self.parent.check_anchors:
                        with target_path.open() as fpt:
                            parsed = html5lib.parse(fpt, namespaceHTMLElements=False)
                            return self.handle_anchor(parsed, anchor)
                    break
            if not exists:
                target_path = dirpath.join(url_path)
                raise BrokenLinkError(url, "No such file: %s" % target_path)


def extensions_str(extensions):
    if not extensions:
        return ''
    extensions = ['"%s"' % e.lstrip('.') for e in extensions if e]
    if len(extensions) == 1:
        return extensions[0]
    return (", ".join(extensions[:-1]) +
            " and %s" % extensions[-1])


def validate_extensions(extensions):
    invalid = set(extensions) - supported_extensions
    if invalid:
        warnings.warn("C1", "Unsupported extensions for check-links: %s" %
            extensions_str(invalid))
